#
# This module is generated with gen_api.pl
# Do not modify.
#

package Net::FreeIPA::API::Data;

use strict;
use warnings;

use version;
our $VERSION = version->new("v2.230");

use Readonly;

# Generated API data using Net::FreeIPA::API::cache
# All commands are JSON encoded.

Readonly our %API_DATA => {
    
    aci_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"permission","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"group","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"filter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"subtree","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"targetgroup","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"selfaci","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":false,"name":"aciprefix","class":"StrEnum","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"test","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"aci_add"}',
    
    aci_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"aciprefix","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"aci_del"}',
    
    aci_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"permission","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"filter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"subtree","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"targetgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"selfaci","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"aciprefix","class":"StrEnum","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"aci_find"}',
    
    aci_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"permission","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"filter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"subtree","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"targetgroup","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"selfaci","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":false,"name":"aciprefix","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"aci_mod"}',
    
    aci_rename => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"permission","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"filter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"subtree","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"targetgroup","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"selfaci","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":false,"name":"aciprefix","class":"StrEnum","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"newname","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"aci_rename"}',
    
    aci_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"aciprefix","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"location","class":"DNParam","type":"DN"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"aci_show"}',
    
    adtrust_is_enabled => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"adtrust_is_enabled"}',
    
    automember_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_add"}',
    
    automember_add_condition => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"automemberinclusiveregex","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"automemberexclusiveregex","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"key","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_add_condition"}',
    
    automember_default_group_remove => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_default_group_remove"}',
    
    automember_default_group_set => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automemberdefaultgroup","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_default_group_set"}',
    
    automember_default_group_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_default_group_show"}',
    
    automember_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_del"}',
    
    automember_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"automember_find"}',
    
    automember_find_orphans => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"remove","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"automember_find_orphans"}',
    
    automember_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_mod"}',
    
    automember_rebuild => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"users","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hosts","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"no_wait","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_rebuild"}',
    
    automember_remove_condition => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"automemberinclusiveregex","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"automemberexclusiveregex","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"key","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_remove_condition"}',
    
    automember_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"type","class":"StrEnum","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automember_show"}',
    
    automountkey_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapautomountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"automountkey","class":"IA5Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountinformation","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountkey_add"}',
    
    automountkey_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapautomountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":false,"name":"automountkey","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"automountinformation","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountkey_del"}',
    
    automountkey_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapautomountmapname","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"automountkey","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"automountinformation","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountkey_find"}',
    
    automountkey_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapautomountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"automountkey","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"automountinformation","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"newautomountinformation","class":"IA5Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","type":"unicode","class":"Str"}],"name":"automountkey_mod"}',
    
    automountkey_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapautomountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":false,"name":"automountkey","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"automountinformation","class":"IA5Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountkey_show"}',
    
    automountlocation_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountlocation_add"}',
    
    automountlocation_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountlocation_del"}',
    
    automountlocation_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"automountlocation_find"}',
    
    automountlocation_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountlocation_show"}',
    
    automountlocation_tofiles => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountlocation_tofiles"}',
    
    automountmap_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountmap_add"}',
    
    automountmap_add_indirect => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":false,"multivalue":false,"name":"key","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"parentmap","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountmap_add_indirect"}',
    
    automountmap_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"automountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountmap_del"}',
    
    automountmap_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"automountmapname","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"automountmap_find"}',
    
    automountmap_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountmap_mod"}',
    
    automountmap_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"automountlocationcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"automountmapname","class":"IA5Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"automountmap_show"}',
    
    batch => '{"takes_args":[{"required":false,"autofill":false,"multivalue":true,"name":"methods","class":"Dict","type":"dict"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"batch"}',
    
    ca_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipacasubjectdn","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"chain","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"ca_add"}',
    
    ca_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"ca_del"}',
    
    ca_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"ca_disable"}',
    
    ca_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"ca_enable"}',
    
    ca_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacaid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacasubjectdn","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacaissuerdn","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"ca_find"}',
    
    ca_is_enabled => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"ca_is_enabled"}',
    
    ca_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"ca_mod"}',
    
    ca_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"chain","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"ca_show"}',
    
    caacl_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacacategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertprofilecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"servicecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"caacl_add"}',
    
    caacl_add_ca => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"ca","class":"Str","type":"unicode"}],"name":"caacl_add_ca"}',
    
    caacl_add_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"caacl_add_host"}',
    
    caacl_add_profile => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"certprofile","class":"Str","type":"unicode"}],"name":"caacl_add_profile"}',
    
    caacl_add_service => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"service","class":"Str","type":"unicode"}],"name":"caacl_add_service"}',
    
    caacl_add_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"caacl_add_user"}',
    
    caacl_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"caacl_del"}',
    
    caacl_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"caacl_disable"}',
    
    caacl_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"caacl_enable"}',
    
    caacl_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacacategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertprofilecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"servicecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"caacl_find"}',
    
    caacl_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacacategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertprofilecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"servicecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"caacl_mod"}',
    
    caacl_remove_ca => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"ca","class":"Str","type":"unicode"}],"name":"caacl_remove_ca"}',
    
    caacl_remove_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"caacl_remove_host"}',
    
    caacl_remove_profile => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"certprofile","class":"Str","type":"unicode"}],"name":"caacl_remove_profile"}',
    
    caacl_remove_service => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"service","class":"Str","type":"unicode"}],"name":"caacl_remove_service"}',
    
    caacl_remove_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"caacl_remove_user"}',
    
    caacl_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"caacl_show"}',
    
    cert_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"certificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":false,"name":"issuer","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"revocation_reason","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cacn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"subject","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"min_serial_number","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"max_serial_number","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"exactly","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"validnotafter_from","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"validnotafter_to","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"validnotbefore_from","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"validnotbefore_to","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"issuedon_from","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"issuedon_to","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"revokedon_from","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"revokedon_to","class":"DateTime","type":"datetime"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"no_host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":true,"name":"no_service","class":"Principal","type":"Principal"}],"name":"cert_find"}',
    
    cert_remove_hold => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"serial_number","type":"int","class":"Int"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"cacn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cert_remove_hold"}',
    
    cert_request => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"csr","class":"CertificateSigningRequest","type":"CertificateSigningRequest"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"request_type","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"profile_id","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"cacn","type":"unicode","class":"Str"},{"required":true,"autofill":false,"multivalue":false,"name":"principal","class":"Principal","type":"Principal"},{"required":true,"autofill":true,"multivalue":false,"name":"add","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"chain","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cert_request"}',
    
    cert_revoke => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"serial_number","type":"int","class":"Int"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"revocation_reason","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"cacn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cert_revoke"}',
    
    cert_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"serial_number","type":"int","class":"Int"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"cacn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"out","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"chain","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"cert_show"}',
    
    cert_status => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"request_id","type":"int","class":"Int"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"cacn","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cert_status"}',
    
    certmap_match => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"certificate","class":"Certificate","type":"Certificate"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmap_match"}',
    
    certmapconfig_mod => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmappromptusername","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmapconfig_mod"}',
    
    certmapconfig_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmapconfig_show"}',
    
    certmaprule_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmapmaprule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmapmatchrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"associateddomain","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmappriority","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmaprule_add"}',
    
    certmaprule_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmaprule_del"}',
    
    certmaprule_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmaprule_disable"}',
    
    certmaprule_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmaprule_enable"}',
    
    certmaprule_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmapmaprule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmapmatchrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"associateddomain","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmappriority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"certmaprule_find"}',
    
    certmaprule_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmapmaprule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmapmatchrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"associateddomain","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertmappriority","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmaprule_mod"}',
    
    certmaprule_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certmaprule_show"}',
    
    certprofile_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certprofile_del"}',
    
    certprofile_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertprofilestoreissued","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"certprofile_find"}',
    
    certprofile_import => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipacertprofilestoreissued","type":"bool","class":"Bool"},{"required":true,"autofill":false,"multivalue":false,"name":"file","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certprofile_import"}',
    
    certprofile_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipacertprofilestoreissued","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"file","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certprofile_mod"}',
    
    certprofile_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"out","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"certprofile_show"}',
    
    class_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"class_find"}',
    
    class_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"full_name","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"class_show"}',
    
    command_defaults => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"full_name","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"params","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"kw","class":"Dict","type":"dict"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"command_defaults"}',
    
    command_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"command_find"}',
    
    command_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"full_name","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"command_show"}',
    
    compat_is_enabled => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"compat_is_enabled"}',
    
    config_mod => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipamaxusernamelength","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipahomesrootdir","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipadefaultloginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipadefaultprimarygroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipadefaultemaildomain","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasearchtimelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasearchrecordslimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipausersearchfields","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipagroupsearchfields","class":"IA5Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipamigrationenabled","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"ipagroupobjectclasses","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserobjectclasses","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapwdexpadvnotify","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"ipaconfigstring","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaselinuxusermaporder","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaselinuxusermapdefault","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipakrbauthzdata","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserauthtype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ca_renewal_master_server","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipadomainresolutionorder","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"config_mod"}',
    
    config_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"config_show"}',
    
    cosentry_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"krbpwdpolicyreference","class":"DNParam","type":"DN"},{"required":true,"autofill":false,"multivalue":false,"name":"cospriority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cosentry_add"}',
    
    cosentry_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cosentry_del"}',
    
    cosentry_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdpolicyreference","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"cospriority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"cosentry_find"}',
    
    cosentry_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdpolicyreference","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"cospriority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cosentry_mod"}',
    
    cosentry_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"cosentry_show"}',
    
    delegation_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"memberof","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"group","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"delegation_add"}',
    
    delegation_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"delegation_del"}',
    
    delegation_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"delegation_find"}',
    
    delegation_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"group","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"delegation_mod"}',
    
    delegation_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"delegation_show"}',
    
    dns_is_enabled => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dns_is_enabled"}',
    
    dns_resolve => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"hostname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dns_resolve"}',
    
    dns_update_system_records => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"dry_run","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dns_update_system_records"}',
    
    dnsconfig_mod => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowsyncptr","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"idnszonerefresh","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipadnsversion","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsconfig_mod"}',
    
    dnsconfig_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsconfig_show"}',
    
    dnsforwardzone_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"name_from_ip","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"skip_overlap_check","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_add"}',
    
    dnsforwardzone_add_permission => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_add_permission"}',
    
    dnsforwardzone_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_del"}',
    
    dnsforwardzone_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_disable"}',
    
    dnsforwardzone_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_enable"}',
    
    dnsforwardzone_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"name_from_ip","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnszoneactive","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"dnsforwardzone_find"}',
    
    dnsforwardzone_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"name_from_ip","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_mod"}',
    
    dnsforwardzone_remove_permission => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_remove_permission"}',
    
    dnsforwardzone_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsforwardzone_show"}',
    
    dnsrecord_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"dnszoneidnsname","type":"DNSName","class":"DNSNameParam"},{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"dnsttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsclass","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"arecord","type":"unicode","class":"ARecord"},{"required":false,"autofill":false,"multivalue":false,"name":"a_part_ip_address","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"a_extra_create_reverse","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"aaaarecord","type":"unicode","class":"AAAARecord"},{"required":false,"autofill":false,"multivalue":false,"name":"aaaa_part_ip_address","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"aaaa_extra_create_reverse","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"a6record","type":"unicode","class":"A6Record"},{"required":false,"autofill":false,"multivalue":false,"name":"a6_part_data","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"afsdbrecord","type":"unicode","class":"AFSDBRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"afsdb_part_subtype","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"afsdb_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"aplrecord","type":"unicode","class":"APLRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"certrecord","type":"unicode","class":"CERTRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_key_tag","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_certificate_or_crl","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"cnamerecord","type":"unicode","class":"CNAMERecord"},{"required":false,"autofill":false,"multivalue":false,"name":"cname_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"dhcidrecord","type":"unicode","class":"DHCIDRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dlvrecord","type":"unicode","class":"DLVRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_key_tag","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_digest_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_digest","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"dnamerecord","type":"unicode","class":"DNAMERecord"},{"required":false,"autofill":false,"multivalue":false,"name":"dname_part_target","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"dsrecord","type":"unicode","class":"DSRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_key_tag","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_digest_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_digest","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"hiprecord","type":"unicode","class":"HIPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ipseckeyrecord","type":"unicode","class":"IPSECKEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"keyrecord","type":"unicode","class":"KEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"kxrecord","type":"unicode","class":"KXRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"kx_part_preference","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"kx_part_exchanger","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"locrecord","type":"unicode","class":"LOCRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_deg","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_min","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_sec","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_dir","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_deg","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_min","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_sec","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_dir","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_altitude","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_size","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_h_precision","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_v_precision","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":true,"name":"mxrecord","type":"unicode","class":"MXRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"mx_part_preference","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"mx_part_exchanger","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"naptrrecord","type":"unicode","class":"NAPTRRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_order","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_preference","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_flags","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_service","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_regexp","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_replacement","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"nsrecord","type":"unicode","class":"NSRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"ns_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"nsecrecord","type":"unicode","class":"NSECRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ptrrecord","type":"unicode","class":"PTRRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"ptr_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"rrsigrecord","type":"unicode","class":"RRSIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"rprecord","type":"unicode","class":"RPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"sigrecord","type":"unicode","class":"SIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"spfrecord","type":"unicode","class":"SPFRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"srvrecord","type":"unicode","class":"SRVRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_priority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_weight","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_port","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_target","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"sshfprecord","type":"unicode","class":"SSHFPRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"sshfp_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sshfp_part_fp_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sshfp_part_fingerprint","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"tlsarecord","type":"unicode","class":"TLSARecord"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_cert_usage","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_selector","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_matching_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_cert_association_data","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"txtrecord","type":"unicode","class":"TXTRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"txt_part_data","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"urirecord","type":"unicode","class":"URIRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"uri_part_priority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"uri_part_weight","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"uri_part_target","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"structured","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsrecord_add"}',
    
    dnsrecord_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"dnszoneidnsname","type":"DNSName","class":"DNSNameParam"},{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"dnsttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsclass","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"arecord","type":"unicode","class":"ARecord"},{"required":false,"autofill":false,"multivalue":true,"name":"aaaarecord","type":"unicode","class":"AAAARecord"},{"required":false,"autofill":false,"multivalue":true,"name":"a6record","type":"unicode","class":"A6Record"},{"required":false,"autofill":false,"multivalue":true,"name":"afsdbrecord","type":"unicode","class":"AFSDBRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"aplrecord","type":"unicode","class":"APLRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"certrecord","type":"unicode","class":"CERTRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"cnamerecord","type":"unicode","class":"CNAMERecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dhcidrecord","type":"unicode","class":"DHCIDRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dlvrecord","type":"unicode","class":"DLVRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dnamerecord","type":"unicode","class":"DNAMERecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dsrecord","type":"unicode","class":"DSRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"hiprecord","type":"unicode","class":"HIPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ipseckeyrecord","type":"unicode","class":"IPSECKEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"keyrecord","type":"unicode","class":"KEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"kxrecord","type":"unicode","class":"KXRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"locrecord","type":"unicode","class":"LOCRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"mxrecord","type":"unicode","class":"MXRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"naptrrecord","type":"unicode","class":"NAPTRRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"nsrecord","type":"unicode","class":"NSRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"nsecrecord","type":"unicode","class":"NSECRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ptrrecord","type":"unicode","class":"PTRRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"rrsigrecord","type":"unicode","class":"RRSIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"rprecord","type":"unicode","class":"RPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"sigrecord","type":"unicode","class":"SIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"spfrecord","type":"unicode","class":"SPFRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"srvrecord","type":"unicode","class":"SRVRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"sshfprecord","type":"unicode","class":"SSHFPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"tlsarecord","type":"unicode","class":"TLSARecord"},{"required":false,"autofill":false,"multivalue":true,"name":"txtrecord","type":"unicode","class":"TXTRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"urirecord","type":"unicode","class":"URIRecord"},{"required":true,"autofill":true,"multivalue":false,"name":"del_all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"structured","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsrecord_del"}',
    
    dnsrecord_delentry => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"dnszoneidnsname","type":"DNSName","class":"DNSNameParam"},{"required":true,"autofill":false,"multivalue":true,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsrecord_delentry"}',
    
    dnsrecord_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"dnszoneidnsname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsclass","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"arecord","type":"unicode","class":"ARecord"},{"required":false,"autofill":false,"multivalue":true,"name":"aaaarecord","type":"unicode","class":"AAAARecord"},{"required":false,"autofill":false,"multivalue":true,"name":"a6record","type":"unicode","class":"A6Record"},{"required":false,"autofill":false,"multivalue":true,"name":"afsdbrecord","type":"unicode","class":"AFSDBRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"aplrecord","type":"unicode","class":"APLRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"certrecord","type":"unicode","class":"CERTRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"cnamerecord","type":"unicode","class":"CNAMERecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dhcidrecord","type":"unicode","class":"DHCIDRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dlvrecord","type":"unicode","class":"DLVRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dnamerecord","type":"unicode","class":"DNAMERecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dsrecord","type":"unicode","class":"DSRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"hiprecord","type":"unicode","class":"HIPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ipseckeyrecord","type":"unicode","class":"IPSECKEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"keyrecord","type":"unicode","class":"KEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"kxrecord","type":"unicode","class":"KXRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"locrecord","type":"unicode","class":"LOCRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"mxrecord","type":"unicode","class":"MXRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"naptrrecord","type":"unicode","class":"NAPTRRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"nsrecord","type":"unicode","class":"NSRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"nsecrecord","type":"unicode","class":"NSECRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ptrrecord","type":"unicode","class":"PTRRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"rrsigrecord","type":"unicode","class":"RRSIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"rprecord","type":"unicode","class":"RPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"sigrecord","type":"unicode","class":"SIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"spfrecord","type":"unicode","class":"SPFRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"srvrecord","type":"unicode","class":"SRVRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"sshfprecord","type":"unicode","class":"SSHFPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"tlsarecord","type":"unicode","class":"TLSARecord"},{"required":false,"autofill":false,"multivalue":true,"name":"txtrecord","type":"unicode","class":"TXTRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"urirecord","type":"unicode","class":"URIRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"structured","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"dnsrecord_find"}',
    
    dnsrecord_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"dnszoneidnsname","type":"DNSName","class":"DNSNameParam"},{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"dnsttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsclass","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"arecord","type":"unicode","class":"ARecord"},{"required":false,"autofill":false,"multivalue":false,"name":"a_part_ip_address","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"aaaarecord","type":"unicode","class":"AAAARecord"},{"required":false,"autofill":false,"multivalue":false,"name":"aaaa_part_ip_address","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"a6record","type":"unicode","class":"A6Record"},{"required":false,"autofill":false,"multivalue":false,"name":"a6_part_data","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"afsdbrecord","type":"unicode","class":"AFSDBRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"afsdb_part_subtype","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"afsdb_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"aplrecord","type":"unicode","class":"APLRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"certrecord","type":"unicode","class":"CERTRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_key_tag","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cert_part_certificate_or_crl","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"cnamerecord","type":"unicode","class":"CNAMERecord"},{"required":false,"autofill":false,"multivalue":false,"name":"cname_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"dhcidrecord","type":"unicode","class":"DHCIDRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"dlvrecord","type":"unicode","class":"DLVRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_key_tag","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_digest_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dlv_part_digest","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"dnamerecord","type":"unicode","class":"DNAMERecord"},{"required":false,"autofill":false,"multivalue":false,"name":"dname_part_target","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"dsrecord","type":"unicode","class":"DSRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_key_tag","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_digest_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ds_part_digest","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"hiprecord","type":"unicode","class":"HIPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ipseckeyrecord","type":"unicode","class":"IPSECKEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"keyrecord","type":"unicode","class":"KEYRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"kxrecord","type":"unicode","class":"KXRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"kx_part_preference","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"kx_part_exchanger","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"locrecord","type":"unicode","class":"LOCRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_deg","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_min","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_sec","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lat_dir","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_deg","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_min","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_sec","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_lon_dir","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_altitude","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_size","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_h_precision","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":false,"name":"loc_part_v_precision","type":"Decimal","class":"Decimal"},{"required":false,"autofill":false,"multivalue":true,"name":"mxrecord","type":"unicode","class":"MXRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"mx_part_preference","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"mx_part_exchanger","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"naptrrecord","type":"unicode","class":"NAPTRRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_order","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_preference","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_flags","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_service","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_regexp","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"naptr_part_replacement","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"nsrecord","type":"unicode","class":"NSRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"ns_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"nsecrecord","type":"unicode","class":"NSECRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"ptrrecord","type":"unicode","class":"PTRRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"ptr_part_hostname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"rrsigrecord","type":"unicode","class":"RRSIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"rprecord","type":"unicode","class":"RPRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"sigrecord","type":"unicode","class":"SIGRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"spfrecord","type":"unicode","class":"SPFRecord"},{"required":false,"autofill":false,"multivalue":true,"name":"srvrecord","type":"unicode","class":"SRVRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_priority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_weight","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_port","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"srv_part_target","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"sshfprecord","type":"unicode","class":"SSHFPRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"sshfp_part_algorithm","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sshfp_part_fp_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sshfp_part_fingerprint","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"tlsarecord","type":"unicode","class":"TLSARecord"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_cert_usage","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_selector","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_matching_type","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"tlsa_part_cert_association_data","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"txtrecord","type":"unicode","class":"TXTRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"txt_part_data","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"urirecord","type":"unicode","class":"URIRecord"},{"required":false,"autofill":false,"multivalue":false,"name":"uri_part_priority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"uri_part_weight","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"uri_part_target","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"structured","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","type":"DNSName","class":"DNSNameParam"}],"name":"dnsrecord_mod"}',
    
    dnsrecord_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"dnszoneidnsname","type":"DNSName","class":"DNSNameParam"},{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"structured","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsrecord_show"}',
    
    dnsrecord_split_parts => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"name","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"value","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsrecord_split_parts"}',
    
    dnsserver_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"idnsserverid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoamname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"dnsserver_find"}',
    
    dnsserver_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsserverid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"idnssoamname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsserver_mod"}',
    
    dnsserver_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsserverid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnsserver_show"}',
    
    dnszone_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"name_from_ip","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoamname","type":"DNSName","class":"DNSNameParam"},{"required":true,"autofill":true,"multivalue":false,"name":"idnssoarname","type":"DNSName","class":"DNSNameParam"},{"required":true,"autofill":true,"multivalue":false,"name":"idnssoaserial","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"idnssoarefresh","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"idnssoaretry","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"idnssoaexpire","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"idnssoaminimum","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsdefaultttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsclass","class":"StrEnum","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"idnsupdatepolicy","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"idnsallowdynupdate","type":"bool","class":"Bool"},{"required":false,"autofill":true,"multivalue":false,"name":"idnsallowquery","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"idnsallowtransfer","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowsyncptr","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssecinlinesigning","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"nsec3paramrecord","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"skip_overlap_check","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"skip_nameserver_check","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"ip_address","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_add"}',
    
    dnszone_add_permission => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_add_permission"}',
    
    dnszone_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_del"}',
    
    dnszone_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_disable"}',
    
    dnszone_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_enable"}',
    
    dnszone_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"name_from_ip","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnszoneactive","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoamname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoarname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaserial","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoarefresh","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaretry","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaexpire","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaminimum","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsdefaultttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsclass","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsupdatepolicy","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowdynupdate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowquery","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowtransfer","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowsyncptr","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssecinlinesigning","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"nsec3paramrecord","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"forward_only","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"dnszone_find"}',
    
    dnszone_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"name_from_ip","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"idnsforwarders","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsforwardpolicy","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoamname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoarname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaserial","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoarefresh","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaretry","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaexpire","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssoaminimum","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsdefaultttl","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"dnsclass","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsupdatepolicy","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowdynupdate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowquery","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowtransfer","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"idnsallowsyncptr","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"idnssecinlinesigning","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"nsec3paramrecord","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_mod"}',
    
    dnszone_remove_permission => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_remove_permission"}',
    
    dnszone_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"dnszone_show"}',
    
    domainlevel_get => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"domainlevel_get"}',
    
    domainlevel_set => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"ipadomainlevel","type":"int","class":"Int"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"domainlevel_set"}',
    
    env => '{"takes_args":[{"multivalue":false,"required":false,"name":"variables","type":"unknown_type","class":"unknown_class","autofill":false}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"server","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"env"}',
    
    group_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"nonposix","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"external","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"group_add"}',
    
    group_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"ipaexternalmember","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"group_add_member"}',
    
    group_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"group_del"}',
    
    group_detach => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"group_detach"}',
    
    group_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"private","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"posix","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"external","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"nonposix","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_sudorule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_sudorule","class":"Str","type":"unicode"}],"name":"group_find"}',
    
    group_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"posix","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"external","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","type":"unicode","class":"Str"}],"name":"group_mod"}',
    
    group_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"ipaexternalmember","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"group_remove_member"}',
    
    group_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"group_show"}',
    
    hbacrule_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"accessruletype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sourcehostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"servicecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacrule_add"}',
    
    hbacrule_add_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"hbacrule_add_host"}',
    
    hbacrule_add_service => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"hbacsvc","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hbacsvcgroup","class":"Str","type":"unicode"}],"name":"hbacrule_add_service"}',
    
    hbacrule_add_sourcehost => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"hbacrule_add_sourcehost"}',
    
    hbacrule_add_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"hbacrule_add_user"}',
    
    hbacrule_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"hbacrule_del"}',
    
    hbacrule_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"hbacrule_disable"}',
    
    hbacrule_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"hbacrule_enable"}',
    
    hbacrule_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"accessruletype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sourcehostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"servicecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"hbacrule_find"}',
    
    hbacrule_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"accessruletype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sourcehostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"servicecategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"hbacrule_mod"}',
    
    hbacrule_remove_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"hbacrule_remove_host"}',
    
    hbacrule_remove_service => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"hbacsvc","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hbacsvcgroup","class":"Str","type":"unicode"}],"name":"hbacrule_remove_service"}',
    
    hbacrule_remove_sourcehost => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"hbacrule_remove_sourcehost"}',
    
    hbacrule_remove_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"hbacrule_remove_user"}',
    
    hbacrule_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacrule_show"}',
    
    hbacsvc_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacsvc_add"}',
    
    hbacsvc_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"hbacsvc_del"}',
    
    hbacsvc_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"hbacsvc_find"}',
    
    hbacsvc_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacsvc_mod"}',
    
    hbacsvc_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacsvc_show"}',
    
    hbacsvcgroup_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacsvcgroup_add"}',
    
    hbacsvcgroup_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"hbacsvc","class":"Str","type":"unicode"}],"name":"hbacsvcgroup_add_member"}',
    
    hbacsvcgroup_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"hbacsvcgroup_del"}',
    
    hbacsvcgroup_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"hbacsvcgroup_find"}',
    
    hbacsvcgroup_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacsvcgroup_mod"}',
    
    hbacsvcgroup_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"hbacsvc","class":"Str","type":"unicode"}],"name":"hbacsvcgroup_remove_member"}',
    
    hbacsvcgroup_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hbacsvcgroup_show"}',
    
    hbactest => '{"takes_args":[],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sourcehost","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"targethost","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"service","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"rules","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"nodetail","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"enabled","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"disabled","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"hbactest"}',
    
    host_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nshostlocation","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nshardwareplatform","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsosversion","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"random","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"macaddress","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaassignedidview","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalauthind","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbrequirespreauth","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbokasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrboktoauthasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"no_reverse","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"ip_address","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"host_add"}',
    
    host_add_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"host_add_cert"}',
    
    host_add_managedby => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"}],"name":"host_add_managedby"}',
    
    host_add_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"host_add_principal"}',
    
    host_allow_create_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"host_allow_create_keytab"}',
    
    host_allow_retrieve_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"host_allow_retrieve_keytab"}',
    
    host_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"updatedns","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"host_del"}',
    
    host_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"host_disable"}',
    
    host_disallow_create_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"host_disallow_create_keytab"}',
    
    host_disallow_retrieve_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"host_disallow_retrieve_keytab"}',
    
    host_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nshostlocation","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nshardwareplatform","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsosversion","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"macaddress","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaassignedidview","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalauthind","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"in_hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_sudorule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_sudorule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"enroll_by_user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_enroll_by_user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"man_by_host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_man_by_host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"man_host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_man_host","class":"Str","type":"unicode"}],"name":"host_find"}',
    
    host_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nshostlocation","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nshardwareplatform","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsosversion","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"random","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":true,"name":"macaddress","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaassignedidview","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalauthind","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbrequirespreauth","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbokasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrboktoauthasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"updatedns","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"host_mod"}',
    
    host_remove_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"host_remove_cert"}',
    
    host_remove_managedby => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"}],"name":"host_remove_managedby"}',
    
    host_remove_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"host_remove_principal"}',
    
    host_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"fqdn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"out","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"host_show"}',
    
    hostgroup_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hostgroup_add"}',
    
    hostgroup_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"hostgroup_add_member"}',
    
    hostgroup_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"hostgroup_del"}',
    
    hostgroup_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"no_host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_sudorule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_sudorule","class":"Str","type":"unicode"}],"name":"hostgroup_find"}',
    
    hostgroup_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hostgroup_mod"}',
    
    hostgroup_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"hostgroup_remove_member"}',
    
    hostgroup_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"hostgroup_show"}',
    
    i18n_messages => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"i18n_messages"}',
    
    idoverridegroup_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idoverridegroup_add"}',
    
    idoverridegroup_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idoverridegroup_del"}',
    
    idoverridegroup_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"idoverridegroup_find"}',
    
    idoverridegroup_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"idoverridegroup_mod"}',
    
    idoverridegroup_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idoverridegroup_show"}',
    
    idoverrideuser_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaoriginaluid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idoverrideuser_add"}',
    
    idoverrideuser_add_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"idoverrideuser_add_cert"}',
    
    idoverrideuser_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idoverrideuser_del"}',
    
    idoverrideuser_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaoriginaluid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"idoverrideuser_find"}',
    
    idoverrideuser_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaoriginaluid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"idoverrideuser_mod"}',
    
    idoverrideuser_remove_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"idoverrideuser_remove_cert"}',
    
    idoverrideuser_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idviewcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaanchoruuid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"fallback_to_ldap","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idoverrideuser_show"}',
    
    idrange_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"ipabaseid","type":"int","class":"Int"},{"required":true,"autofill":false,"multivalue":false,"name":"ipaidrangesize","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipabaserid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasecondarybaserid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainsid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"iparangetype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idrange_add"}',
    
    idrange_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idrange_del"}',
    
    idrange_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipabaseid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaidrangesize","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipabaserid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasecondarybaserid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainsid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"iparangetype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"idrange_find"}',
    
    idrange_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipabaseid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaidrangesize","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipabaserid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasecondarybaserid","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainsid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainname","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idrange_mod"}',
    
    idrange_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idrange_show"}',
    
    idview_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipadomainresolutionorder","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idview_add"}',
    
    idview_apply => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idview_apply"}',
    
    idview_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idview_del"}',
    
    idview_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"idview_find"}',
    
    idview_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipadomainresolutionorder","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"idview_mod"}',
    
    idview_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"show_hosts","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idview_show"}',
    
    idview_unapply => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"idview_unapply"}',
    
    join => '{"takes_args":[{"required":true,"autofill":true,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"realm","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nshardwareplatform","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsosversion","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"join"}',
    
    json_metadata => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"objname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"methodname","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"object","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"method","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"command","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"json_metadata"}',
    
    kra_is_enabled => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"kra_is_enabled"}',
    
    krbtpolicy_mod => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"uid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"krbmaxticketlife","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbmaxrenewableage","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"krbtpolicy_mod"}',
    
    krbtpolicy_reset => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"uid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"krbtpolicy_reset"}',
    
    krbtpolicy_show => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"uid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"krbtpolicy_show"}',
    
    location_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"location_add"}',
    
    location_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"location_del"}',
    
    location_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"location_find"}',
    
    location_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"location_mod"}',
    
    location_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"idnsname","type":"DNSName","class":"DNSNameParam"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"location_show"}',
    
    migrate_ds => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"ldapuri","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"bindpw","class":"Password","type":"unicode"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"binddn","class":"DNParam","type":"DN"},{"required":true,"autofill":true,"multivalue":false,"name":"usercontainer","class":"DNParam","type":"DN"},{"required":true,"autofill":true,"multivalue":false,"name":"groupcontainer","class":"DNParam","type":"DN"},{"required":true,"autofill":true,"multivalue":true,"name":"userobjectclass","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":true,"name":"groupobjectclass","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":true,"name":"userignoreobjectclass","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":true,"name":"userignoreattribute","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":true,"name":"groupignoreobjectclass","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":true,"name":"groupignoreattribute","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"groupoverwritegid","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"schema","type":"unicode","class":"StrEnum"},{"required":false,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"basedn","class":"DNParam","type":"DN"},{"required":false,"autofill":true,"multivalue":false,"name":"compat","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"cacertfile","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"use_def_group","type":"bool","class":"Bool"},{"required":true,"autofill":true,"multivalue":false,"name":"scope","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":true,"name":"exclude_groups","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":true,"name":"exclude_users","type":"unicode","class":"Str"}],"name":"migrate_ds"}',
    
    netgroup_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nisdomainname","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"netgroup_add"}',
    
    netgroup_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"netgroup","class":"Str","type":"unicode"}],"name":"netgroup_add_member"}',
    
    netgroup_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"netgroup_del"}',
    
    netgroup_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nisdomainname","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"ipauniqueid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"private","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"managed","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_user","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"no_host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"no_hostgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_netgroup","type":"unicode","class":"Str"}],"name":"netgroup_find"}',
    
    netgroup_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nisdomainname","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"netgroup_mod"}',
    
    netgroup_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"netgroup","class":"Str","type":"unicode"}],"name":"netgroup_remove_member"}',
    
    netgroup_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"netgroup_show"}',
    
    otpconfig_mod => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipatokentotpauthwindow","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokentotpsyncwindow","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenhotpauthwindow","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenhotpsyncwindow","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"otpconfig_mod"}',
    
    otpconfig_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"otpconfig_show"}',
    
    otptoken_add => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenuniqueid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"type","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenowner","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokendisabled","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokennotbefore","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokennotafter","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenvendor","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenmodel","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenserial","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"ipatokenotpkey","class":"OTPTokenKey","type":"str"},{"required":false,"autofill":true,"multivalue":false,"name":"ipatokenotpalgorithm","type":"unicode","class":"StrEnum"},{"required":false,"autofill":true,"multivalue":false,"name":"ipatokenotpdigits","type":"int","class":"IntEnum"},{"required":false,"autofill":true,"multivalue":false,"name":"ipatokentotpclockoffset","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"ipatokentotptimestep","type":"int","class":"Int"},{"required":false,"autofill":true,"multivalue":false,"name":"ipatokenhotpcounter","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"qrcode","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"no_qrcode","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"otptoken_add"}',
    
    otptoken_add_managedby => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"ipatokenuniqueid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"}],"name":"otptoken_add_managedby"}',
    
    otptoken_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"ipatokenuniqueid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"otptoken_del"}',
    
    otptoken_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenuniqueid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenowner","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokendisabled","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokennotbefore","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokennotafter","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenvendor","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenmodel","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenserial","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenotpalgorithm","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenotpdigits","type":"int","class":"IntEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokentotpclockoffset","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokentotptimestep","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenhotpcounter","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"otptoken_find"}',
    
    otptoken_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"ipatokenuniqueid","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenowner","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokendisabled","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokennotbefore","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokennotafter","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenvendor","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenmodel","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenserial","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"otptoken_mod"}',
    
    otptoken_remove_managedby => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"ipatokenuniqueid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"}],"name":"otptoken_remove_managedby"}',
    
    otptoken_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"ipatokenuniqueid","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"otptoken_show"}',
    
    output_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"commandfull_name","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"output_find"}',
    
    output_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"commandfull_name","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"name","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"output_show"}',
    
    param_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"metaobjectfull_name","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"param_find"}',
    
    param_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"metaobjectfull_name","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"name","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"param_show"}',
    
    passwd => '{"takes_args":[{"required":true,"autofill":true,"multivalue":false,"name":"principal","class":"Principal","type":"Principal"},{"required":true,"autofill":false,"multivalue":false,"name":"password","class":"Password","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"current_password","class":"Password","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"otp","class":"Password","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"passwd"}',
    
    permission_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"ipapermright","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"ipapermbindruletype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermlocation","class":"DNOrURL","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"extratargetfilter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermtargetfilter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtarget","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtargetto","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtargetfrom","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"targetgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"filter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"subtree","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"permission_add"}',
    
    permission_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"privilege","class":"Str","type":"unicode"}],"name":"permission_add_member"}',
    
    permission_add_noaci => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":false,"multivalue":true,"name":"ipapermissiontype","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"permission_add_noaci"}',
    
    permission_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"permission_del"}',
    
    permission_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermright","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermincludedattr","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermexcludedattr","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermdefaultattr","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermbindruletype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermlocation","class":"DNOrURL","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"extratargetfilter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermtargetfilter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtarget","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtargetto","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtargetfrom","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"targetgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"filter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"subtree","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"permission_find"}',
    
    permission_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"ipapermright","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermincludedattr","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermexcludedattr","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermbindruletype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermlocation","class":"DNOrURL","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"extratargetfilter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipapermtargetfilter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtarget","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtargetto","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"ipapermtargetfrom","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"memberof","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"targetgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"type","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"filter","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"subtree","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","type":"unicode","class":"Str"}],"name":"permission_mod"}',
    
    permission_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"privilege","class":"Str","type":"unicode"}],"name":"permission_remove_member"}',
    
    permission_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"permission_show"}',
    
    ping => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"ping"}',
    
    pkinit_status => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"server_server","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"status","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"pkinit_status"}',
    
    plugins => '{"takes_args":[],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"server","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"plugins"}',
    
    privilege_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"privilege_add"}',
    
    privilege_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"role","class":"Str","type":"unicode"}],"name":"privilege_add_member"}',
    
    privilege_add_permission => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"permission","class":"Str","type":"unicode"}],"name":"privilege_add_permission"}',
    
    privilege_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"privilege_del"}',
    
    privilege_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"privilege_find"}',
    
    privilege_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"privilege_mod"}',
    
    privilege_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"role","class":"Str","type":"unicode"}],"name":"privilege_remove_member"}',
    
    privilege_remove_permission => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"permission","class":"Str","type":"unicode"}],"name":"privilege_remove_permission"}',
    
    privilege_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"privilege_show"}',
    
    pwpolicy_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"krbmaxpwdlife","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbminpwdlife","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdhistorylength","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdmindiffchars","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdminlength","type":"int","class":"Int"},{"required":true,"autofill":false,"multivalue":false,"name":"cospriority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdmaxfailure","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdfailurecountinterval","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdlockoutduration","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"pwpolicy_add"}',
    
    pwpolicy_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"pwpolicy_del"}',
    
    pwpolicy_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"krbmaxpwdlife","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbminpwdlife","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdhistorylength","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdmindiffchars","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdminlength","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cospriority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdmaxfailure","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdfailurecountinterval","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdlockoutduration","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"pwpolicy_find"}',
    
    pwpolicy_mod => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"krbmaxpwdlife","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbminpwdlife","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdhistorylength","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdmindiffchars","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdminlength","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"cospriority","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdmaxfailure","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdfailurecountinterval","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpwdlockoutduration","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"pwpolicy_mod"}',
    
    pwpolicy_show => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"user","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"pwpolicy_show"}',
    
    radiusproxy_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipatokenradiusserver","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"ipatokenradiussecret","class":"Password","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiustimeout","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusretries","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenusermapattribute","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"radiusproxy_add"}',
    
    radiusproxy_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"radiusproxy_del"}',
    
    radiusproxy_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusserver","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiussecret","class":"Password","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiustimeout","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusretries","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenusermapattribute","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"radiusproxy_find"}',
    
    radiusproxy_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusserver","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiussecret","class":"Password","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiustimeout","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusretries","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenusermapattribute","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"radiusproxy_mod"}',
    
    radiusproxy_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"radiusproxy_show"}',
    
    realmdomains_mod => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"associateddomain","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"add_domain","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"del_domain","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"realmdomains_mod"}',
    
    realmdomains_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"realmdomains_show"}',
    
    role_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"role_add"}',
    
    role_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"service","class":"Str","type":"unicode"}],"name":"role_add_member"}',
    
    role_add_privilege => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"privilege","class":"Str","type":"unicode"}],"name":"role_add_privilege"}',
    
    role_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"role_del"}',
    
    role_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"role_find"}',
    
    role_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"role_mod"}',
    
    role_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"service","class":"Str","type":"unicode"}],"name":"role_remove_member"}',
    
    role_remove_privilege => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"privilege","class":"Str","type":"unicode"}],"name":"role_remove_privilege"}',
    
    role_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"role_show"}',
    
    schema => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"known_fingerprints","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"schema"}',
    
    selfservice_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selfservice_add"}',
    
    selfservice_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selfservice_del"}',
    
    selfservice_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"aciname","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selfservice_find"}',
    
    selfservice_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"permissions","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"attrs","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selfservice_mod"}',
    
    selfservice_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"aciname","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selfservice_show"}',
    
    selinuxusermap_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"ipaselinuxuser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"seealso","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"selinuxusermap_add"}',
    
    selinuxusermap_add_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"selinuxusermap_add_host"}',
    
    selinuxusermap_add_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"selinuxusermap_add_user"}',
    
    selinuxusermap_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selinuxusermap_del"}',
    
    selinuxusermap_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selinuxusermap_disable"}',
    
    selinuxusermap_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"selinuxusermap_enable"}',
    
    selinuxusermap_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaselinuxuser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"seealso","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"selinuxusermap_find"}',
    
    selinuxusermap_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipaselinuxuser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"seealso","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"selinuxusermap_mod"}',
    
    selinuxusermap_remove_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"selinuxusermap_remove_host"}',
    
    selinuxusermap_remove_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"selinuxusermap_remove_user"}',
    
    selinuxusermap_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"selinuxusermap_show"}',
    
    server_conncheck => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"remote_cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"server_conncheck"}',
    
    server_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"ignore_topology_disconnect","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"ignore_last_of_role","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"server_del"}',
    
    server_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipamindomainlevel","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"ipamaxdomainlevel","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"topologysuffix","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"no_topologysuffix","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_location","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_location","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":true,"name":"servrole","class":"Str","type":"unicode"}],"name":"server_find"}',
    
    server_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipalocation_location","type":"DNSName","class":"DNSNameParam"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaserviceweight","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"server_mod"}',
    
    server_role_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"server_server","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"role_servrole","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"status","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"include_master","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"server_role_find"}',
    
    server_role_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"server_server","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"role_servrole","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"server_role_show"}',
    
    server_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"server_show"}',
    
    service_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"ipakrbauthzdata","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalauthind","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbrequirespreauth","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbokasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrboktoauthasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"force","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"service_add"}',
    
    service_add_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"service_add_cert"}',
    
    service_add_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"}],"name":"service_add_host"}',
    
    service_add_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"},{"required":true,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"service_add_principal"}',
    
    service_allow_create_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"service_allow_create_keytab"}',
    
    service_allow_retrieve_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"service_allow_retrieve_keytab"}',
    
    service_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"service_del"}',
    
    service_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"service_disable"}',
    
    service_disallow_create_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"service_disallow_create_keytab"}',
    
    service_disallow_retrieve_keytab => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"}],"name":"service_disallow_retrieve_keytab"}',
    
    service_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":true,"name":"ipakrbauthzdata","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalauthind","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"man_by_host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_man_by_host","class":"Str","type":"unicode"}],"name":"service_find"}',
    
    service_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"ipakrbauthzdata","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalauthind","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbrequirespreauth","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrbokasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"ipakrboktoauthasdelegate","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"service_mod"}',
    
    service_remove_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"service_remove_cert"}',
    
    service_remove_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"}],"name":"service_remove_host"}',
    
    service_remove_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"},{"required":true,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"service_remove_principal"}',
    
    service_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"krbcanonicalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"out","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"service_show"}',
    
    servicedelegationrule_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"servicedelegationrule_add"}',
    
    servicedelegationrule_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"principal","class":"Str","type":"unicode"}],"name":"servicedelegationrule_add_member"}',
    
    servicedelegationrule_add_target => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"servicedelegationtarget","class":"Str","type":"unicode"}],"name":"servicedelegationrule_add_target"}',
    
    servicedelegationrule_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"servicedelegationrule_del"}',
    
    servicedelegationrule_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"servicedelegationrule_find"}',
    
    servicedelegationrule_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"principal","class":"Str","type":"unicode"}],"name":"servicedelegationrule_remove_member"}',
    
    servicedelegationrule_remove_target => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"servicedelegationtarget","class":"Str","type":"unicode"}],"name":"servicedelegationrule_remove_target"}',
    
    servicedelegationrule_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"servicedelegationrule_show"}',
    
    servicedelegationtarget_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"servicedelegationtarget_add"}',
    
    servicedelegationtarget_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"principal","class":"Str","type":"unicode"}],"name":"servicedelegationtarget_add_member"}',
    
    servicedelegationtarget_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"servicedelegationtarget_del"}',
    
    servicedelegationtarget_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"servicedelegationtarget_find"}',
    
    servicedelegationtarget_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"principal","class":"Str","type":"unicode"}],"name":"servicedelegationtarget_remove_member"}',
    
    servicedelegationtarget_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"servicedelegationtarget_show"}',
    
    session_logout => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"session_logout"}',
    
    sidgen_was_run => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"sidgen_was_run"}',
    
    stageuser_activate => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"stageuser_activate"}',
    
    stageuser_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"givenname","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"sn","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"displayname","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"initials","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":false,"name":"krbprincipalexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpasswordexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":true,"name":"mail","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","type":"unicode","class":"Password"},{"required":false,"autofill":true,"multivalue":false,"name":"random","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"street","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"st","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"postalcode","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"telephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"mobile","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"pager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"facsimiletelephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ou","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"title","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"manager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"carlicense","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserauthtype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusconfiglink","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiususername","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"departmentnumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeetype","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"preferredlanguage","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"from_delete","type":"bool","class":"Bool"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"stageuser_add"}',
    
    stageuser_add_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"stageuser_add_cert"}',
    
    stageuser_add_certmapdata => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"ipacertmapdata","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"issuer","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"subject","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"certificate","class":"Certificate","type":"Certificate"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"stageuser_add_certmapdata"}',
    
    stageuser_add_manager => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"}],"name":"stageuser_add_manager"}',
    
    stageuser_add_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"stageuser_add_principal"}',
    
    stageuser_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"stageuser_del"}',
    
    stageuser_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"givenname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"displayname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"initials","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":false,"name":"krbprincipalexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpasswordexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":true,"name":"mail","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","type":"unicode","class":"Password"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"street","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"st","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"postalcode","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"telephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"mobile","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"pager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"facsimiletelephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ou","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"title","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"manager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"carlicense","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserauthtype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusconfiglink","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiususername","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"departmentnumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeetype","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"preferredlanguage","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"in_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_sudorule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_sudorule","class":"Str","type":"unicode"}],"name":"stageuser_find"}',
    
    stageuser_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"givenname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"displayname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"initials","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":false,"name":"krbprincipalexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpasswordexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":true,"name":"mail","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","type":"unicode","class":"Password"},{"required":false,"autofill":true,"multivalue":false,"name":"random","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"street","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"st","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"postalcode","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"telephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"mobile","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"pager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"facsimiletelephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ou","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"title","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"manager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"carlicense","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserauthtype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusconfiglink","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiususername","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"departmentnumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeetype","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"preferredlanguage","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","type":"unicode","class":"Str"}],"name":"stageuser_mod"}',
    
    stageuser_remove_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"stageuser_remove_cert"}',
    
    stageuser_remove_certmapdata => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"ipacertmapdata","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"issuer","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"subject","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"certificate","class":"Certificate","type":"Certificate"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"stageuser_remove_certmapdata"}',
    
    stageuser_remove_manager => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"}],"name":"stageuser_remove_manager"}',
    
    stageuser_remove_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"stageuser_remove_principal"}',
    
    stageuser_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"stageuser_show"}',
    
    sudocmd_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"sudocmd","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudocmd_add"}',
    
    sudocmd_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"sudocmd","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"sudocmd_del"}',
    
    sudocmd_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"sudocmd","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"sudocmd_find"}',
    
    sudocmd_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"sudocmd","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudocmd_mod"}',
    
    sudocmd_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"sudocmd","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudocmd_show"}',
    
    sudocmdgroup_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudocmdgroup_add"}',
    
    sudocmdgroup_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmd","class":"Str","type":"unicode"}],"name":"sudocmdgroup_add_member"}',
    
    sudocmdgroup_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"sudocmdgroup_del"}',
    
    sudocmdgroup_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"sudocmdgroup_find"}',
    
    sudocmdgroup_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudocmdgroup_mod"}',
    
    sudocmdgroup_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmd","class":"Str","type":"unicode"}],"name":"sudocmdgroup_remove_member"}',
    
    sudocmdgroup_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudocmdgroup_show"}',
    
    sudorule_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cmdcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasusercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasgroupcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sudoorder","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"externaluser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasextuser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasextgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudorule_add"}',
    
    sudorule_add_allow_command => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmd","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmdgroup","class":"Str","type":"unicode"}],"name":"sudorule_add_allow_command"}',
    
    sudorule_add_deny_command => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmd","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmdgroup","class":"Str","type":"unicode"}],"name":"sudorule_add_deny_command"}',
    
    sudorule_add_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostmask","class":"Str","type":"unicode"}],"name":"sudorule_add_host"}',
    
    sudorule_add_option => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"ipasudoopt","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudorule_add_option"}',
    
    sudorule_add_runasgroup => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"sudorule_add_runasgroup"}',
    
    sudorule_add_runasuser => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"sudorule_add_runasuser"}',
    
    sudorule_add_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"sudorule_add_user"}',
    
    sudorule_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"sudorule_del"}',
    
    sudorule_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"sudorule_disable"}',
    
    sudorule_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"sudorule_enable"}',
    
    sudorule_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cmdcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasusercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasgroupcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sudoorder","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"externaluser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasextuser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasextgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"sudorule_find"}',
    
    sudorule_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipaenabledflag","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"usercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"hostcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cmdcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasusercategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasgroupcategory","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sudoorder","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"externaluser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"externalhost","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasextuser","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipasudorunasextgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","class":"Str","type":"unicode"}],"name":"sudorule_mod"}',
    
    sudorule_remove_allow_command => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmd","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmdgroup","class":"Str","type":"unicode"}],"name":"sudorule_remove_allow_command"}',
    
    sudorule_remove_deny_command => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmd","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"sudocmdgroup","class":"Str","type":"unicode"}],"name":"sudorule_remove_deny_command"}',
    
    sudorule_remove_host => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"host","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostgroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"hostmask","class":"Str","type":"unicode"}],"name":"sudorule_remove_host"}',
    
    sudorule_remove_option => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"ipasudoopt","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudorule_remove_option"}',
    
    sudorule_remove_runasgroup => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"sudorule_remove_runasgroup"}',
    
    sudorule_remove_runasuser => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"sudorule_remove_runasuser"}',
    
    sudorule_remove_user => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"}],"name":"sudorule_remove_user"}',
    
    sudorule_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"sudorule_show"}',
    
    topic_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"topic_find"}',
    
    topic_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"full_name","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topic_show"}',
    
    topologysegment_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"topologysuffixcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"iparepltoposegmentleftnode","type":"unicode","class":"Str"},{"required":true,"autofill":false,"multivalue":false,"name":"iparepltoposegmentrightnode","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"iparepltoposegmentdirection","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicastripattrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatedattributelist","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatedattributelisttotal","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatimeout","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicaenabled","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysegment_add"}',
    
    topologysegment_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"topologysuffixcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysegment_del"}',
    
    topologysegment_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"topologysuffixcn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"iparepltoposegmentleftnode","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"iparepltoposegmentrightnode","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"iparepltoposegmentdirection","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicastripattrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatedattributelist","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatedattributelisttotal","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatimeout","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicaenabled","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"topologysegment_find"}',
    
    topologysegment_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"topologysuffixcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicastripattrs","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatedattributelist","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatedattributelisttotal","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicatimeout","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"nsds5replicaenabled","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysegment_mod"}',
    
    topologysegment_reinitialize => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"topologysuffixcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":true,"multivalue":false,"name":"left","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"right","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"stop","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysegment_reinitialize"}',
    
    topologysegment_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"topologysuffixcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysegment_show"}',
    
    topologysuffix_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"iparepltopoconfroot","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysuffix_add"}',
    
    topologysuffix_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysuffix_del"}',
    
    topologysuffix_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"iparepltopoconfroot","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"topologysuffix_find"}',
    
    topologysuffix_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"iparepltopoconfroot","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysuffix_mod"}',
    
    topologysuffix_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysuffix_show"}',
    
    topologysuffix_verify => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"topologysuffix_verify"}',
    
    trust_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"trust_type","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"realm_admin","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"realm_passwd","class":"Password","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"realm_server","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"trust_secret","class":"Password","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"base_id","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"range_size","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"range_type","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"bidirectional","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"external","type":"bool","class":"Bool"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trust_add"}',
    
    trust_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trust_del"}',
    
    trust_fetch_domains => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"realm_server","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trust_fetch_domains"}',
    
    trust_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipantflatname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainsid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipantsidblacklistincoming","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipantsidblacklistoutgoing","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"trust_find"}',
    
    trust_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":true,"name":"ipantsidblacklistincoming","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipantsidblacklistoutgoing","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipantadditionalsuffixes","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trust_mod"}',
    
    trust_resolve => '{"takes_args":[],"takes_options":[{"required":true,"autofill":false,"multivalue":true,"name":"sids","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trust_resolve"}',
    
    trust_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trust_show"}',
    
    trustconfig_mod => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipantfallbackprimarygroup","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"trust_type","type":"unicode","class":"StrEnum"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trustconfig_mod"}',
    
    trustconfig_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"trust_type","type":"unicode","class":"StrEnum"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trustconfig_show"}',
    
    trustdomain_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"trustcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipantflatname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainsid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"trust_type","type":"unicode","class":"StrEnum"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trustdomain_add"}',
    
    trustdomain_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"trustcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":true,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trustdomain_del"}',
    
    trustdomain_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"trustcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trustdomain_disable"}',
    
    trustdomain_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"trustcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trustdomain_enable"}',
    
    trustdomain_find => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"trustcn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipantflatname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainsid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"trustdomain_find"}',
    
    trustdomain_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"trustcn","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"ipantflatname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipanttrusteddomainsid","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"trust_type","type":"unicode","class":"StrEnum"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"trustdomain_mod"}',
    
    user_add => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":false,"multivalue":false,"name":"givenname","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"sn","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"displayname","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"initials","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":false,"name":"krbprincipalexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpasswordexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":true,"name":"mail","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","type":"unicode","class":"Password"},{"required":false,"autofill":true,"multivalue":false,"name":"random","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"street","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"st","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"postalcode","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"telephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"mobile","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"pager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"facsimiletelephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ou","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"title","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"manager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"carlicense","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserauthtype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusconfiglink","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiususername","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"departmentnumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeetype","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"preferredlanguage","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":false,"name":"nsaccountlock","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"noprivate","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"user_add"}',
    
    user_add_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"user_add_cert"}',
    
    user_add_certmapdata => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"ipacertmapdata","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"issuer","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"subject","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"certificate","class":"Certificate","type":"Certificate"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"user_add_certmapdata"}',
    
    user_add_manager => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"}],"name":"user_add_manager"}',
    
    user_add_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"user_add_principal"}',
    
    user_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"preserve","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"user_del"}',
    
    user_disable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"user_disable"}',
    
    user_enable => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"user_enable"}',
    
    user_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"givenname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"displayname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"initials","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":false,"name":"krbprincipalexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpasswordexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":true,"name":"mail","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","type":"unicode","class":"Password"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"street","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"st","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"postalcode","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"telephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"mobile","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"pager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"facsimiletelephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ou","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"title","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"manager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"carlicense","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserauthtype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusconfiglink","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiususername","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"departmentnumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeetype","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"preferredlanguage","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":false,"name":"nsaccountlock","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"preserved","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":true,"autofill":true,"multivalue":false,"name":"whoami","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"in_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_group","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_netgroup","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_role","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_hbacrule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"in_sudorule","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"not_in_sudorule","class":"Str","type":"unicode"}],"name":"user_find"}',
    
    user_mod => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"givenname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"sn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"cn","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"displayname","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"initials","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"homedirectory","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"gecos","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"loginshell","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"},{"required":false,"autofill":false,"multivalue":false,"name":"krbprincipalexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":false,"name":"krbpasswordexpiration","class":"DateTime","type":"datetime"},{"required":false,"autofill":false,"multivalue":true,"name":"mail","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"userpassword","type":"unicode","class":"Password"},{"required":false,"autofill":true,"multivalue":false,"name":"random","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"uidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"gidnumber","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"street","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"l","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"st","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"postalcode","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"telephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"mobile","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"pager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"facsimiletelephonenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ou","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"title","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"manager","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"carlicense","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipasshpubkey","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"ipauserauthtype","class":"StrEnum","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"userclass","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiusconfiglink","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipatokenradiususername","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"departmentnumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeenumber","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"employeetype","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"preferredlanguage","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"},{"required":false,"autofill":false,"multivalue":false,"name":"nsaccountlock","type":"bool","class":"Bool"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"rename","type":"unicode","class":"Str"}],"name":"user_mod"}',
    
    user_remove_cert => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":true,"autofill":false,"multivalue":true,"name":"usercertificate","class":"Certificate","type":"Certificate"}],"name":"user_remove_cert"}',
    
    user_remove_certmapdata => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"ipacertmapdata","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"issuer","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":false,"name":"subject","class":"DNParam","type":"DN"},{"required":false,"autofill":false,"multivalue":true,"name":"certificate","class":"Certificate","type":"Certificate"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"user_remove_certmapdata"}',
    
    user_remove_manager => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"}],"name":"user_remove_manager"}',
    
    user_remove_principal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":true,"name":"krbprincipalname","class":"Principal","type":"Principal"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"user_remove_principal"}',
    
    user_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"out","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"user_show"}',
    
    user_stage => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"user_stage"}',
    
    user_status => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"useruid","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"user_status"}',
    
    user_undel => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"user_undel"}',
    
    user_unlock => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"uid","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"user_unlock"}',
    
    vault_add_internal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"ipavaulttype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"ipavaultsalt","class":"Bytes","type":"str"},{"required":false,"autofill":false,"multivalue":false,"name":"ipavaultpublickey","class":"Bytes","type":"str"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"vault_add_internal"}',
    
    vault_add_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"services","class":"Str","type":"unicode"}],"name":"vault_add_member"}',
    
    vault_add_owner => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"services","class":"Str","type":"unicode"}],"name":"vault_add_owner"}',
    
    vault_archive_internal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"session_key","class":"Bytes","type":"str"},{"required":true,"autofill":false,"multivalue":false,"name":"vault_data","class":"Bytes","type":"str"},{"required":true,"autofill":false,"multivalue":false,"name":"nonce","class":"Bytes","type":"str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"vault_archive_internal"}',
    
    vault_del => '{"takes_args":[{"required":true,"autofill":false,"multivalue":true,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"vault_del"}',
    
    vault_find => '{"takes_args":[{"required":false,"autofill":false,"multivalue":false,"name":"criteria","class":"Str","type":"unicode"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipavaulttype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"timelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"sizelimit","type":"int","class":"Int"},{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":false,"autofill":true,"multivalue":false,"name":"services","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"users","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":true,"multivalue":false,"name":"pkey_only","type":"bool","class":"Flag"}],"name":"vault_find"}',
    
    vault_mod_internal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"description","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"ipavaulttype","type":"unicode","class":"StrEnum"},{"required":false,"autofill":false,"multivalue":false,"name":"ipavaultsalt","class":"Bytes","type":"str"},{"required":false,"autofill":false,"multivalue":false,"name":"ipavaultpublickey","class":"Bytes","type":"str"},{"required":false,"autofill":false,"multivalue":true,"name":"setattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"addattr","type":"unicode","class":"Str"},{"required":false,"autofill":false,"multivalue":true,"name":"delattr","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"vault_mod_internal"}',
    
    vault_remove_member => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"services","class":"Str","type":"unicode"}],"name":"vault_remove_member"}',
    
    vault_remove_owner => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"services","class":"Str","type":"unicode"}],"name":"vault_remove_owner"}',
    
    vault_retrieve_internal => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":false,"multivalue":false,"name":"session_key","class":"Bytes","type":"str"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"vault_retrieve_internal"}',
    
    vault_show => '{"takes_args":[{"required":true,"autofill":false,"multivalue":false,"name":"cn","type":"unicode","class":"Str"}],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"vault_show"}',
    
    vaultconfig_show => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"transport_out","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"vaultconfig_show"}',
    
    vaultcontainer_add_owner => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"services","class":"Str","type":"unicode"}],"name":"vaultcontainer_add_owner"}',
    
    vaultcontainer_del => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"continue","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"vaultcontainer_del"}',
    
    vaultcontainer_remove_owner => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":true,"name":"user","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"group","class":"Str","type":"unicode"},{"required":false,"autofill":false,"multivalue":true,"name":"services","class":"Str","type":"unicode"}],"name":"vaultcontainer_remove_owner"}',
    
    vaultcontainer_show => '{"takes_args":[],"takes_options":[{"required":true,"autofill":true,"multivalue":false,"name":"rights","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"service","class":"Principal","type":"Principal"},{"required":false,"autofill":true,"multivalue":false,"name":"shared","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"username","class":"Str","type":"unicode"},{"required":true,"autofill":true,"multivalue":false,"name":"all","type":"bool","class":"Flag"},{"required":true,"autofill":true,"multivalue":false,"name":"raw","type":"bool","class":"Flag"},{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"},{"required":true,"autofill":true,"multivalue":false,"name":"no_members","type":"bool","class":"Flag"}],"name":"vaultcontainer_show"}',
    
    whoami => '{"takes_args":[],"takes_options":[{"required":false,"autofill":false,"multivalue":false,"name":"version","type":"unicode","class":"Str"}],"name":"whoami"}',

};

1;
